#!/usr/bin/env python
# Copyright (C) 2006 Kumar McMillan
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

"""tools for automatically triggering test suites.

... for example, a continuous integeration test triggered 
from a `subversion post-commit hook`_.

Since repositories in real life contain several packages that most likely 
will not be tested in the exact same way, there is no generic script 
provided here to trigger test suites.

Defining your own bgtest script is easy though. 
Here is an example with custom packages ::

    from testtools.bgtests import TestablePackage, run_packages
    class MyPackage(TestablePackage):
        svn_path = 'file://%s/widget/trunk'
        ex = 'nosetests -v'
        def recognizes_file(self, filepath, code=None):
            return filepath.startswith('widget/trunk')
        
    run_packages( [MyPackage()], '/path/to/repos', notify='egg@yourface' )

You can fire something like this off very nicely inside a `subversion post-commit hook`_   
by passing through the repos_root and revision.

The default test command is `nosetests`_, a discovery test runner for python, 
but any command can be used here.

If you want to run your tests on a machine separate from that which holds your
repository, then you can run a SimpleXMLRPCServer using the `PackageTestHandler` 
(see its docstring for an example).

.. _subversion post-commit hook: http://svnbook.red-bean.com/nightly/en/svn-book.html#svn.reposadmin.create.hooks
.. _nosetests: http://nose.python-hosting.com/

"""

import os
import sys
import sre
import optparse
import inspect
from tempfile import mkdtemp
from datetime import datetime
from subprocess import Popen, PIPE, STDOUT
import smtplib
import logging
from testtools.io import TempIO
try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO
from SimpleXMLRPCServer import SimpleXMLRPCServer
from warnings import warn


class TestablePackage(object):
    """encapsulates a package that can be extracted and tested.
    
    """
    ex = 'nosetests'
    svnpath = None # URL for this package's contents, including protocol
    
    def __init__(self, **kw):
        [setattr(self,k,v) for k,v in kw.items()]
        if self.svnpath is None:
            raise ValueError, "svnpath cannot be None"
    
    def get_name(self):
        """get the name of this package.
        
        looks for self.__name__ or else uses the class name of self.
        
        """
        if hasattr(self,'__name__'):
            return self.__name__
        else:
            return self.__class__.__name__
    
    def skip_notify(self, buffer):
        """when True, notification is skipped.
        
        it's only called on error.  
        buffer is the output of the failing test(s)
        
        """
        return False
    
    def recognizes_file(self, filepath, code=None):
        """should return True if filepath from repos affects this package.
        
        filepath is grabbed from the output of svnlook.
        it starts with the first dir relative to the repository root 
        and will not contain a forward slash.            
            
        """
        return False
        
    def runtests(   self, rev=None, stream=sys.stdout, 
                    notify=None, header='', tmp_dir=None):
        """runs tests for this package.
        
        returns True if tests passed, false otherwise.
        
        Keyword Arguments:
        ------------------
        - rev -- optional revision to export (head otherwise)
        - stream -- filelike object for reporter, defaults to sys.stdout
        - notify -- if not empty, results are emailed here
        - header -- string to prepend onto test suite output
        - tmp_dir -- directory to export into and run tests (i.e. /tmp, the default)        
    
        """
        reporter = Reporter(stream=stream, notify=notify, 
                            header=header, skip_notify=self.skip_notify )
        res = run(  self.svnpath, reporter, rev = rev, test_cmd = self.ex,
                    tmp_prefix=self.get_name()+'_', tmp_dir=tmp_dir,
                    setup = self.setup )
        
        return res
    
    def setup(self):
        """hook for package to setup environment, etc. before running tests.
        
        """
        pass

class Reporter(object):
    """reports the test results.
    
    outputs to a log and/or email address.
    
    Keyword Arguments:
    ------------------
    - stream -- filelike object for reporter
    - notify -- if not empty, results are emailed here
    - export -- svn path we are reporting about
    - revision -- revision we are reporting about
    - header -- gets added to buffer
    - skip_notify -- callback that accepts buffer contents on error and \
                    returns True if notification should be skipped
    - em_from -- email From address
    
    """
    def __init__(   self, stream=sys.stdout, notify=None, 
                    export='unknown', revision='unknown', header='',
                    skip_notify=None, em_from='bgtests@localhost'):
        self.export = export
        self.revision = revision
        self.stream = stream
        self.buffer = StringIO()
        if header:
            self.write(header)
        self.notify = notify
        self.em_from = em_from
        
        if callable(skip_notify):
            self._skip_notify = skip_notify
        else:
            def sk(buf): return False
            self._skip_notify = sk
            
    def finish(self, tests_passed=False):
        if not tests_passed and self.notify:
            output = self.buffer.getvalue()
            if self._skip_notify(output):
                print >>self, " ** notification skipped"
                return
                
            print >>self, " ** emailing error report to %s" % self.notify
            
            s = smtplib.SMTP('localhost')
            msg = "\n".join([   "From: %s" % self.em_from, "To: %s" % self.notify,
                                "Subject: bgtests FAILED: %s:%s" % ( self.export, 
                                                                    self.revision),
                                "", output])
                        
            s.sendmail(self.em_from, self.notify, msg)
            s.quit()
    
    def start(self, export=None, revision=None, tester=None):
        def override(k,v):
            if v is not None: setattr(self, k,v)
        
        override('export', export)
        override('revision', revision)
        override('tester', tester)
        
        print >>self, '[start: %s]'         % datetime.now()
        print >>self, '[svn_path: %s]'      % self.export
        print >>self, "[svn_revision: %s]"  % self.revision
        print >>self, "[tester: %s]"        % tester
    
    def write(self, chunk):
        self.stream.write(chunk)
        self.buffer.write(chunk)

class PackageTestHandler(object):
    """runs test packages via XMLRPC.
    
    a handler designed for `SimpleXMLRPCServer`.
    
    Example of starting the server::
    
        pkg = TestablePackage(svnpath = 'http://svn.myrepo/')
        handler = PackageTestHandler([pkg])
        serv = SimpleXMLRPCServer(("localhost", 8000))
        serv.register_introspection_functions()
        serv.register_instance(handler)
        serv.serve_forever()
    
    You send a request to the server and call a method to trigger tests.
    see `trigger_from_svnlook` for a concrete example of client/server interaction.
    
    Arguments:
    ----------
    - packages -- list of `TestablePackage` objects
    
    Keyword Arguments:
    ------------------
    - stream -- filelike object for reporter, defaults to sys.stdout
    - notify -- if not empty, results are emailed here
    - tmp_dir -- directory to export into and run tests
                        
    """
    def __init__(   self, packages, stream=sys.stdout, 
                    notify=None, tmp_dir=None ):
        self.packages = packages
        self.stream = stream
        self.notify = notify
        self.tmp_dir = tmp_dir
        
    def trigger_from_svnlook(self, svnlook_info, svnlook_changed, revision):
        """analyzes input from `svnlook changed` and runs tests affected by the change.
        
        this method is designed to support the scenario where your subversion \
        repository is located on one machine but you want to run the tests on another.
        
        A client that calls this from within a `subversion post-commit hook`_ 
        might look like this::
            
            from testtools.bgtests import SvnInspector
            from xmlrpclib import ServerProxy
            meta = SvnInspector(repos_root, revision=revision)
            s = ServerProxy('http://%s:%s' % (host, port))
            passed = s.trigger_from_svnlook( meta.info(), "\\n".join(meta.changed()), revision )

        NOTE: the handler assumes the svnlook output is from the same repository 
        that the packages reside in and currently doesn't not support mixing packages that 
        come from different repositories.
        
        returns True if all tests passed (or if none ran), False otherwise.

        .. _subversion post-commit hook: http://svnbook.red-bean.com/nightly/en/svn-book.html#svn.reposadmin.create.hooks
        
        """
        class repos_meta:
            info = StringIO(svnlook_info).getvalue
            changed = StringIO(svnlook_changed).readlines
            
        passed = run_packages(  self.packages, rev=revision, repos_meta=repos_meta,
                                stream=self.stream, notify=self.notify, tmp_dir=self.tmp_dir)
        self.stream.flush() # when running as daemon
        return passed

class SvnInspector:
    """encapsulates meta data for a repository.
    
    currently this is subversion specific.
    
    Arguments:
    ----------
    - repos_root -- abs path to base of repository (cannot be a URL)
    
    Keyword Arguments:
    ------------------
    - revision -- inspect this revision (defaults to head)
                    
    """
    def __init__(self, repos_root, revision=None):
        self.repos_root = repos_root
        self.revision = revision
    
    def changed(self):
        """yields each line in the `svnlook changed` command.
        
        returns an iterable.
        
        """
        cmd = 'svnlook changed %s' % self.repos_root
        if self.revision is not None:
            cmd = cmd + ' --revision %s' % self.revision
        
        cmd = Popen(cmd, stdout=PIPE, shell=True, env=os.environ)
        for line in cmd.stdout.readlines():
            yield line
        
        r = cmd.wait()
        if r != 0:
            raise RuntimeError, '`svnlook changed` exited: %s' % r
    
    def info(self):
        """get `svnlook info` command output.
        
        returns a string.
        
        """
        cmd = ['svnlook','info']
        if self.revision is not None:
            cmd.extend( ['--revision', str(self.revision)] )
            
        cmd.append(self.repos_root)
        return Popen( cmd, stdout=PIPE, stderr=STDOUT).communicate()[0]

def run(svn_path, out=None, rev=None, test_cmd='nosetests', stream=sys.stdout, 
        notify=None, tmp_dir=None, tmp_prefix='tmp_bgtest_', setup=None):
    """runs tests in the background.
    
    exports svn_path (URL to package) into a temp dir and runs test_cmd.
    
    Keyword Arguments:
    ------------------
    - out -- reporter object, will be created otherwise
    - rev -- optional revision to export (head otherwise)
    - test_cmd -- command to run tests with
    - stream -- filelike object for reporter, defaults to sys.stdout
    - notify -- if not empty, results are emailed here
    - tmp_dir -- directory to export into and run tests
    - tmp_prefix -- a unique temporary directory is created with this prefix
    - setup -- a callback to do any setup work once you are *in* the dir of your project
    
    returns True if test passed, false otherwise.
    
    """
    if out is None:
        out = Reporter( stream=stream, notify=notify )
        
    pkg_dir = TempIO(dir=tmp_dir, prefix=tmp_prefix)
    cwd = os.getcwd()
    try:
    
        os.chdir(pkg_dir.root)
        
        # force export into dir because we just created an empty dir ...
        cmd = 'svn export "%s" . --force -q' % (svn_path) 
        if rev is not None:
            cmd = cmd + ' -r %s' % rev
        
        svn = Popen(cmd, env=os.environ, shell=True)
        svn.wait()
        assert svn.returncode == 0, 'svn exited %s; cmd="%s"' % (svn.returncode, cmd)
        
        if setup is not None:
            setup()
            
        out.start(export=svn_path, revision=rev, tester=test_cmd)
        
        # not using stderr=STDOUT anymore because we need
        # to always expect stderr to be at the bottom of the report (i.e. for cavendish)
        tester = Popen(test_cmd, stdout=PIPE, stderr=PIPE, 
                            shell=True, env=os.environ)

        for line in tester.stdout.readlines():
            out.write(line)
        for line in tester.stderr.readlines():
            out.write(line)
            
        r = tester.wait()
        test_passed = (r == 0)
        out.finish( test_passed )
        
    finally:
        os.chdir(cwd)
        del pkg_dir
        
    return test_passed

def run_packages(   packages, repos_root=None, rev=None, stream=sys.stdout, 
                    notify=None, tmp_dir=None, repos_meta=None):
    """runs tests for packages who were affected by the last change in a repository.
    
    gets last changed paths from a repos_meta object to ask
    each package if it was affected by any one of the file paths.
    If so, the package's test suite is `run` in the manner defined by the package.
    
    Arguments:
    ----------
    - packages -- list of `TestablePackage` objects
    
    Keyword Arguments:
    ------------------
    - repos_root -- abs path to base of repository (cannot be a URL); \
                    may be omitted in lieu of a repos_meta object
    - rev -- revision to inspect, defaults to None (head of repos)
    - stream -- filelike object for reporter, defaults to sys.stdout
    - notify -- if not empty, results are emailed here
    - tmp_dir -- directory to export into and run tests
    - repos_meta -- instance that provides the interface of `SvnInspector`
    
    see `run` for how svn exporting works.
    
    returns True if all tests passed (or none ran), False otherwise.
    
    """
    if repos_meta is None and repos_root is None:
        raise ValueError, "repos_root cannot be None in absence of a repos_meta object"
    if repos_meta is None:
        repos_meta = SvnInspector(repos_root = repos_root, revision = rev)
    
    # the incriminating info :
    header = repos_meta.info()
    
    svnpath = None
    statline = sre.compile(r'([^\s]+)\s+(.*)')
    pkgs_to_test = []
    
    for line in repos_meta.changed():
        line = line.strip()
        if line == '':
            continue
        m = statline.match(line)
        if m is None:
            # I wonder if this should be an exception.  or a warning.  hmm
            print >>stream, " ** unparsable line in `svnlook changed`: '%s'" % line
            continue
        else:
            code, file = m.groups()
        
        for pk in packages:
            if pk.recognizes_file(file, code=code):
                if pk not in pkgs_to_test:
                    pkgs_to_test.append(pk)
    
    if len(pkgs_to_test) == 0:
        print >>stream, " ** no tests to run for repos=%s, revision=%s" % (repos_root, rev)
        return True
    
    all_passed = True
    for pkg in pkgs_to_test:
        passed = pkg.runtests(  rev=rev, stream=stream, notify=notify, 
                                header=header, tmp_dir=tmp_dir)
        if not passed:
            all_passed = False
    
    return all_passed
