#!/usr/bin/env python

"""generate testtools fixture from a data source"""

## TODO: the SO logic should be in
## SOStor and this should be more generic so that any SOStor object 
## can generate a fixture from a data source.

import sys
import os
import optparse
import pprint
import re
import inspect
import md5
from sqlobject.col import SOForeignKey

class code_str(str):
    """string that reproduces without quotes.
    
    """
    def __repr__(self):
        return str.__repr__(self)[1:-1]

class SOFixtureGenerator(object):
    """builds a callable fixture-code generator.
    
    """
    template = """
class %(fxt_class)s(SOFixture):
    class meta(basemeta):
        so_class = %(so_class)s
    def data(self):
        %(data_header)s
        
        return %(data)s
"""
    def __init__(self, context, limit=None, so_class_trans={}):
        self.context = context
        self.limit = limit
        self.so_class_trans = so_class_trans
        self.fxt_code = []
        self.fxt_code_map = {}
        self.header = {}
        
    def __call__( self, model, clauseTables=[], 
                query=None, outfile=None, orderBy=None, 
                show_query_only=False ):
        """generates fixture code from settings.
        
        """
        if query is not None:
            rs = model.select(query, clauseTables=clauseTables, 
                                                orderBy=orderBy)
        else:
            rs = model.select(orderBy=orderBy)
        if show_query_only:
            print rs
            return
        
        self.to_fxt_code(rs, model )
        dupeless_code = [ c for c in self.fxt_code if c != "" ]
        
        # somebody slap me (so that I use a proper template already) :
        code = ("\n".join(self.header.keys()) + 
                    "\n\nclass basemeta:\n    pass\n" +
                    ("\n".join(dupeless_code)))
        
        if outfile is not None:
            if os.path.exists(outfile):
                raise OSError("file '%s' already exists" % outfile)
            f = open(outfile, 'w')
            f.write(code)
            f.close()
    
            print "wrote fixture to file %s" % outfile
        else:
            print code
    
    def add_fxt_code(self, fxt_class, code):
        """register a chunk of code for a `Fixture` class.
        
        """
        # we need to weed out duplicates but keep the
        # top-most duplicate so that the cascade will work
        if self.fxt_code_map.has_key(fxt_class):
            return
            
        self.fxt_code.append(code)
        i = len(self.fxt_code)-1
        self.fxt_code_map[fxt_class] = i
    
    def gen_fxt_cls(self, model_name):
        """return the name for a `Fixture` class.
        
        """
        return self.context.capitalize() + '_' + model_name
    
    def gen_fxt_id(self, table, counter):
        """return an index suitable for a `Fixture` dataset.
        
        """
        return "_".join( [self.context, table, str(counter)] )
    
    def to_fxt_code(self, rs, model, descending=0 ):
    
        if not descending and self.limit is not None:
            # we want unlimited foreign key values 
            # (with a limit on the parent query this should be ok 
            #  unless your table has a million fkeys ;) )
            rs = rs[0:self.limit]
        
        table = model.sqlmeta.table
        style = model.sqlmeta.style
        columnList = model.sqlmeta.columnList
        columns = model.sqlmeta.columns
        model_name = model.__name__
    
        fkeys = [] # fkeys in order of appearance
        fkey_idmap = {} # col.name : list of ids
        fkey_classmap = {}
        for name,col in columns.items():
            if isinstance(col, SOForeignKey):
                dbcol = style.pythonAttrToDBColumn(col.name)
                #print [a for a in dir(col) if not a.startswith('_')]
                fkey_classmap[dbcol] = col.foreignKey
                
                # translate derived soClass name into a real one 
                # (if available and if necessary)
                newmodel_name = fkey_classmap[dbcol]
                if self.so_class_trans.has_key(newmodel_name):
                    fkey_classmap[dbcol] = self.so_class_trans[newmodel_name]
                    
                fkey_idmap[dbcol] = []
    
        fixture=[]
        data_header={}
        count = 1
        def getDbName(col):
            if col.dbName is not None:
                return col.dbName
            else:
                return style.pythonAttrToDBColumn(col.name)
                
        for row in rs:
            cols = [style.idForTable(table)]
            cols.extend([getDbName(c) for c in columnList])
        
            vals = [row.id]
            vals.extend([getattr(row, c.name) for c in columnList])
        
            rowmap = dict(zip(cols, vals))
            for col,val in rowmap.items():
                if not val:
                    # since we are only looking for f. keys
                    # if id is empty, we don't care about linking ...
                    continue
                if fkey_idmap.has_key(col):
                    if col not in fkeys:
                        fkeys.append(col)
                    try:
                        index = fkey_idmap[col].index(val) + 1 # 0->1, etc
                    except ValueError:
                        fkey_idmap[col].append(val) # push the ID
                        index = len(fkey_idmap[col])
                    
                    # use fixture val ...
                    # i.e. 
                    # self.meta.req.fkey = Contact()
                    # fkey.metris_contact_1.id
                    # NOTE: see the orderBy clause below for why the
                    #  index works 
                    fxtclass = self.gen_fxt_cls(fkey_classmap[col])
                    fxtvar = style.pythonAttrToDBColumn(fkey_classmap[col])
                    data_header['self.meta.req.%s = %s()' % (
                                                fxtvar, fxtclass)] = 1
                    rowmap[col] = code_str( 'r.%s.%s.id' % \
                                            (fxtvar, self.gen_fxt_id( 
                                                            fxtvar, index )))
    
            fixture.append( (self.gen_fxt_id( 
                            style.pythonAttrToDBColumn(model_name), count), 
                                                                    rowmap) )
            count += 1
        
        # finish ...
        data_header = data_header.keys()
        data_header.append('r = self.meta.req')
        fixture = tuple(fixture)
        self.header['import datetime'] = 1
        self.header['from testtools.fixtures import SOFixture'] = 1
        # import the SO :
        self.header['from %s import %s' % (model.__module__, model_name)] = 1
        
        if sys.version_info[0] >= 2 and sys.version_info[1] >= 4:
            data = pprint.pformat(fixture, indent=3)
        else:
            data = pprint.pformat(fixture)
        code = self.template % dict(fxt_class = self.gen_fxt_cls(model_name),
                                    so_class = model_name,
                                    data = data,
                                    data_header = ("\n" + (
                                                    " "*8)).join(data_header))
        
        for f in fkeys:
            newmodel_name = fkey_classmap[f]
            m = __import__(model.__module__, globals(), locals(), 
                                                [newmodel_name]) 
            newmodel = getattr(m, newmodel_name)
            
            ids = [str(i) for i in fkey_idmap[f] if i is not None]
            if ids:
                ord = []
                for id in ids:
                    # need to preserve order of fkey_idmap so
                    # doing a weird on/off switch in order by
                    ord.append( newmodel.q.id != id )
                            
                rs = newmodel.select('id in (%s)' % ','.join(ids), 
                                                        orderBy=ord)
                self.to_fxt_code(rs, newmodel, descending=descending+1)
        
        self.add_fxt_code(self.gen_fxt_cls(model_name), code)

def main(so_class_trans = {}):
    """Generates fixture code from a real data source.
    
    """
    parser = optparse.OptionParser(
        usage=('%prog [options] context_key package.model.ModelName' 
                                        + "\n\n" + inspect.getdoc(main)))
    parser.add_option('-q','--query',
                help="like, \"id = 1705\" ")
    parser.add_option('--show_query_only', action='store_true',
                help="prints out query generated by sqlobject and exits")
    parser.add_option('-c','--clause_tables', default=[],
                help="comma separated list of tables for query")
    parser.add_option('-l','--limit', type='int',
                help="max results to return")
    parser.add_option('-o','--outfile', metavar='FILE',
                help="output to file instead of stdout")
    parser.add_option('-s','--order_by',
                help="orderBy=ORDER_BY")
    parser.add_option('-d','--dsn',
                help="sets sqlmeta.processConnection")
    
    (options, args) = parser.parse_args()
    try:
        context, model_path = args
    except ValueError:
        parser.error('incorrect arguments')
    
    name, model_name = os.path.splitext(model_path)
    if not model_name:
        model = __import__(name, globals(), locals(), [])
    else:
        if model_name.startswith('.'):
            model_name = model_name[1:]
        model = __import__(name, globals(), locals(), [model_name]) 
        model = getattr(model, model_name)
    
    if options.dsn:
        from sqlobject import sqlhub, connectionForURI
        sqlhub.processConnection = connectionForURI(options.dsn)
    
    generate = SOFixtureGenerator(context, limit=options.limit, 
                                    so_class_trans=so_class_trans )
    
    if options.clause_tables:
        options.clause_tables = options.clause_tables.split(',')
        
    generate( model, show_query_only=options.show_query_only,
            clauseTables=options.clause_tables, query=options.query, 
            outfile=options.outfile, orderBy=options.order_by)

# OUCH! refactor me (code in genfix and genmodel)
# prepare for sqlobject monkey patch :( ...
# this is so that foreign key lookups work right when 
# there are multiple schemas having the same table 
# (perfectly legal, but sqlobject only finds the first primary key)
import re
def columnsFromSchema(self, tableName, soClass):

    keyQuery = """
    SELECT pg_catalog.pg_get_constraintdef(oid) as condef
    FROM pg_catalog.pg_constraint r
    WHERE r.conrelid = %s::regclass AND r.contype = 'f'"""

    colQuery = """
    SELECT a.attname,
    pg_catalog.format_type(a.atttypid, a.atttypmod), a.attnotnull,
    (SELECT substring(d.adsrc for 128) FROM pg_catalog.pg_attrdef d
    WHERE d.adrelid=a.attrelid AND d.adnum = a.attnum)
    FROM pg_catalog.pg_attribute a
    WHERE a.attrelid =%s::regclass
    AND a.attnum > 0 AND NOT a.attisdropped
    ORDER BY a.attnum"""
    
    # kumar: add limit 1 to get primary key for first rel in schema search path
    primaryKeyQuery = """
    SELECT pg_index.indisprimary,
        pg_catalog.pg_get_indexdef(pg_index.indexrelid)
    FROM pg_catalog.pg_class c, pg_catalog.pg_class c2,
        pg_catalog.pg_index AS pg_index
    WHERE c.relname = %s
        AND c.oid = pg_index.indrelid
        AND pg_index.indexrelid = c2.oid
        AND pg_index.indisprimary
    LIMIT 1
    """

    keyData = self.queryAll(keyQuery % self.sqlrepr(tableName))
    keyRE = re.compile(r"\((.+)\) REFERENCES (.+)\(")
    keymap = {}

    for (condef,) in keyData:
        match = keyRE.search(condef)
        if match:
            field, reftable = match.groups()
            keymap[field] = reftable.capitalize()

    primaryData = self.queryAll(primaryKeyQuery % self.sqlrepr(tableName))
    primaryRE = re.compile(r'CREATE .*? USING .* \((.+?)\)')
    primaryKey = None
    for isPrimary, indexDef in primaryData:
        match = primaryRE.search(indexDef)
        assert match, "Unparseable contraint definition: %r" % indexDef
        assert primaryKey is None, "Already found primary key (%r), then found: %r" % (primaryKey, indexDef)
        primaryKey = match.group(1)
    assert primaryKey, "No primary key found in table %r" % tableName
    if primaryKey.startswith('"'):
        assert primaryKey.endswith('"')
        primaryKey = primaryKey[1:-1]

    colData = self.queryAll(colQuery % self.sqlrepr(tableName))
    results = []
    if self.unicodeCols:
        client_encoding = self.queryOne("SHOW client_encoding")[0]
    for field, t, notnull, defaultstr in colData:
        if field == primaryKey:
            continue
        colClass, kw = self.guessClass(t)
        if self.unicodeCols and colClass == col.StringCol:
            colClass = col.UnicodeCol
            kw['dbEncoding'] = client_encoding
        kw['name'] = soClass.sqlmeta.style.dbColumnToPythonAttr(field)
        kw['dbName'] = field
        kw['notNone'] = notnull
        if defaultstr is not None:
            kw['default'] = self.defaultFromSchema(colClass, defaultstr)
        elif not notnull:
            kw['default'] = None
        if keymap.has_key(field):
            kw['foreignKey'] = keymap[field]
        results.append(colClass(**kw))
    return results
from sqlobject.postgres import pgconnection
pgconnection.PostgresConnection.columnsFromSchema = columnsFromSchema
    
if __name__ == '__main__':
    main()