# Copyright (C) 2006 Kumar McMillan
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

"""Fixture storage media.

You most likely will only need the wrapper classes in `fixtures`.

"""

import csv
import sys
try:
    from elementtree.SimpleXMLWriter import XMLWriter
except ImportError:
    XMLWriter = None
from exceptions import *
from config import debug

class Stor(object):
    """interface for a storage medium."""
    def clean(self):
        """erases any previous contents in storage."""
        raise NotImplementedError
    def close(self):
        """perform any close logic, like closing a file."""
        raise NotImplementedError
    def save(self, **kwds):
        """saves fields in kwds."""
        raise NotImplementedError

class SOStor(Stor):
    """SQLObject storage medium.
    
    Arguments
    ---------
    
    so_class: the SQLObject class to actually stor with 
    
    Keyword Arguments
    -----------------
    
    autocommit -- when False (the default) everything is performed within \
        a transaction until `SOStor.clean` is called.  When True, \
        everything is still within a transaction but it is committed \
        upon `SOStor.save`.  Autcommit mode seems to be necessary to work \
        around instances where postgresql would deadlock if you are working \
        with lots of foreign keyed fixtures within transactions
    
    """
    _trans_per_conn = {} # transactions need to be singletons per 
                         # dsn, so that we can work with multiple objects 
                         # (i.e. foreign keys)
                         # and ... we need transactions to work around 
                         # issues with sqlite in sqlobject
                         
    def __init__(self, so_class, autocommit=False):
        from sqlobject.styles import getStyle
        from sqlobject.dbconnection import Transaction
        
        self.style = getStyle(so_class)
        self.so_class = so_class
        self.autocommit = autocommit
        
        cls = self.__class__
        conn = self.so_class._connection
        if (not hasattr(conn, 'dsn') and
                hasattr(conn, 'filename')):
            # sqlite ..
            connid = conn.filename
        else:
            # postgres and ..?..
            connid = conn.dsn
        
        if not cls._trans_per_conn.has_key(connid):
            cls._trans_per_conn[connid] = conn.transaction()
        self._trans = cls._trans_per_conn[connid]
    
    def __repr__(self):
        return "<class '%s' so='%s' at %s>" % \
                        (self.__class__.__name__, 
                         self.so_class.__name__, 
                         hex(id(self)))
    
    def clean(self):
        self.so_class.clearTable()
    
    def close(self):
        self._trans.commit()
        
    def save(self, **kw):
        k = [self.style.dbColumnToPythonAttr(k) for k in kw.keys()]
        v = kw.values()
        new_kw = dict(zip(k,v))
        new_kw['connection'] = self._trans
        self.so_class(**new_kw)
        if self.autocommit:
            self._trans.commit()
            try:
                self._trans.begin()
            except AssertionError:
                # if we are in a bad transaction
                # we should just swallow the "transaction is _obsolete" error?
                pass

class FileStor(Stor):
    """abstract Stor object that involves storing a file.
    
    """
    def __init__(self, file):
        if type(file) is type(""):
            self.file = open(file, 'w')
        else:
            self.file = file
    
    def _new_file(self, fname):
        return open(fname, 'w')
    
    def clean(self):
        # is there a better way ?  truncate() no worky
        fname = self.file.name
        self.file.close()
        self.file = self._new_file(fname)
    
    def close(self):
        self.file.close()

class CsvStor(csv.DictWriter, FileStor):
    """CSV storage medium.
    
    """
    def __init__(self, file, fields, addfields=False, 
                                encoding=None, clean=False, dialect='excel'):
        """needs a filename or file-like object and list of fields.
        
        if encoding is set, all values will be encoded with this type
        before values are saved
        
        """
        FileStor.__init__(self, file)
        self.encoding = encoding
        self.addfields = addfields
        self.fields = fields
        self.dialect = dialect
        if clean:
            FileStor.clean(self)
        self._init_csv()
    
    def _init_csv(self):
        csv.DictWriter.__init__(self, self.file, 
                                    self.fields, dialect=self.dialect)
        
        if self.addfields:
            self.writerow(dict([(f,f) for f in self.fields]))
    
    def clean(self):
        FileStor.clean(self)
        self._init_csv()
        
    def save(self, **kwds):
        if self.encoding:
            kwds = dict( 
                  [(k.encode(self.encoding), 
                    v.encode(self.encoding)) for k,v in kwds.items()] )
        self.writerow(kwds)

class XmlStor(FileStor):
    """XML file storage medium.
    
    note that this is not meant to get crazy, but just
    meant to make an XML representation of *tabular* data ...
    in other words the structure needs to be basically: 
    
        <rows><row><col>...</col></row></rows>
    
    """
    
    def __init__(self, file, encoding='utf-8', item='item', root='dataset'):
        if XMLWriter is None:
            # force an ImportError
            import elementtree.SimpleXMLWriter.XMLWriter
        
        FileStor.__init__(self, file)
        # note that 'us-ascii' is elementtree default
        self.item = item
        self.root = root
        self.encoding = encoding
        self.xml = XMLWriter(self.file, self.encoding)
        self._started = False
        self._root_elem = None
    
    def close(self):
        self.xml.close(self._root_elem)
        FileStor.close(self)
    
    def clean(self):
        FileStor.clean(self)
        # hmm, still not sure about this file-cleaning approach
        self.xml = XMLWriter(self.file, self.encoding)
        self.xml.declaration()
        self._started = False
        self._root_elem = None
    
    def save(self, **kw):
        if not self._started:
            self._root_elem = self.xml.start(self.root)
            self._started = True
            
        self.start_save()
        for k,v in kw.items():
            self.xml.element(k, str(v).encode(self.encoding))
        self.end_save()
    
    def start_save(self):
        self.xml.start(self.item)
        
    def end_save(self):
        self.xml.end(self.item)
    