# Copyright (C) 2006 Kumar McMillan
# 
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.

"""Tools for replacing objects with stubs.

There is not very much here yet because Python allows you to do a lot of stubbing 
naturally.  The `Stub` class provides a few shortcuts though.

"""

from copy import copy
import weakref
from types import *

class NoValue: pass

class Stub(object):
    """Replace an object with a stub object
    
    Let's face it, creating stub objects is dead simple in Python.  
    For example, a simple way to create a stub of foo is :
    
        >>> def foo():
        ...     return 'foo'
        >>> def newfoo():
        ...     return 'bar'
        >>> oldfoo = foo
        >>> try:
        ...     foo = newfoo
        ...     foo()
        ... finally:
        ...     foo = oldfoo
        'bar'
    
    With `Stub`, this is still possible, like so, and it doesn't save you much coding either:
    
        >>> def foo():
        ...     return 'foo'
        >>> def newfoo():
        ...     return 'bar'
        >>> try:
        ...     foo = Stub(foo, replace=newfoo)
        ...     foo()
        ... finally:
        ...     foo = foo.restore()
        'bar'
    
    However, you can do more complex things too, like mimic sequential calls:
    
        >>> def next_ingredient():
        ...     for i in ('eggs', 'flour'):
        ...         yield i
        >>> orig_id = id(next_ingredient)
        >>> next_ingredient = Stub(next_ingredient)
        >>> def yeast():
        ...     return 'yeast'
        >>> def butter():
        ...     return 'butter'
        >>> try:
        ...     next_ingredient.replace([yeast, butter])
        ...     assert next_ingredient() == 'yeast'
        ...     assert next_ingredient() == 'butter'
        ...     next_ingredient()
        ... finally:
        ...     next_ingredient = next_ingredient.restore()
        Traceback (most recent call last):
            ...
        StopIteration
        >>> assert id(next_ingredient) == orig_id
    
    To accomplish this without `Stub` you would have to proxy a generator, 
    which would be just a bit more coding.
    
    In closing ... you probably don't need stub for simple stubbing but it
    might come in handy.  The only reason I added it to testtools is because I had
    some tests that needed urlopen() to send a sequence of responses back in a test.
    
    """
    class IterableCallProxy:
        def __init__(self, calls, getcall=None):
            self.calls = calls
            if getcall is None:
                getcall = iter(calls)
            self.getcall = getcall
            
        def __call__(self, *args, **kw):
            call = self.getcall.next()
            return call(*args, **kw)
    
    def __init__(self, fn, replace=NoValue):
        self.original = fn
        def nocall(*args, **kw):
            raise ValueError, "no stub call has been set for fn %s" % self.original
        self.call_proxy = nocall
        
        if replace is not NoValue:
            self.replace(replace)
    
    def __call__(self, *args, **kw):
        return self.call_proxy(*args, **kw)
    
    def replace(self, new):
        """Replace stub with an iterable of calls or a single callable"""
        try:
            getcall = iter(new)
        except TypeError:
            if not callable(new):
                raise TypeError, "call replacement is not iterable and is not callable"
            self.call_proxy = new
        else:
            self.call_proxy = Stub.IterableCallProxy(new, getcall=getcall)
    
    def restore(self):
        """returns original object for id"""
        return self.original


class stub_template:
    class _class:
        def __init__(*a,**kw): pass
    class _object(object):
        def __init__(*a,**kw): pass
    def _def_unbound(self, *a,**kw): pass
    ## because it's not re-bound when copied to another class obj.
    ## really this doesn't do anything anyway, and can be overidden
    _def_unbound = staticmethod(_def_unbound)
        
def _def(*a,**kw): pass
stub_template._def = _def
    
def mkinterface(cls, template=stub_template):
    """make a class having the same interface as cls.
    
    every method accepts any input but does nothing.
    returns a class object.
    
    """
    if issubclass(cls, object):
        newcls = template._object
    else:
        newcls = template._class
    for name in dir(cls):
        if name.startswith('__'):
            # not sure if we need to go this far
            continue
            
        attr = getattr(cls,name)
        atype = type(attr)
        #print name, attr, atype
        
        # should we make copies instead?
        if atype in (FunctionType, LambdaType, GeneratorType):
            setattr(newcls, name, template._def)
        elif atype == UnboundMethodType:
            setattr(newcls, name, template._def_unbound)
        elif atype == ClassType:
            setattr(newcls, name, template._class)
        elif atype in (StringType, IntType, LongType, FloatType, BooleanType):
            # ok to reference ?
            setattr(newcls, name, attr)
        elif atype in (TypeType, ObjectType):
            setattr(newcls, name, template._object)
        else:
            # ok to forget about it?
            setattr(newcls, name, None)
    
    newcls.__name__ = cls.__name__                     
    return newcls
    