
import unittest
import csv
import os
from os import path

from testtools import config
from testtools.config import reset_debug, set_debug_stream
from testtools.exceptions import *
from testtools.io import TempIO, readfile
from testtools import fixtures
from testtools.fixtures import *
from testtools.fixtures.registry import LoadMap
from testtools.fixtures import registry
from testtools.storage import *

from testtools.testself.util import start_csv
from testtools.testself.myfixtures import EmployeeData, DepartmentData, ShopData, csvdict
from testtools.testself import myfixtures
from testtools.testself import model

TMP = None

def setup_module(self):
    self.TMP = TempIO()
    
def teardown_module(self):
    del self.TMP

data = (('blue_hair_color_employee', {
                    'id':1, 'name':'Fred', 'hair_color':'blue'}),
        ('boring_employee', {
                    'id':2, 'name':'Bob', 'hair_color':'gray'}))
            
class TestCsvFixture(unittest.TestCase):
    
    def setUp(self):
        LoadMap.clear()

    def test_fixture(self):
        fields = ['id', 'name', 'hair_color']
        csvfile = TMP.mkfile('test_fixtures.csv')
        
        # write a row to make sure file gets cleaned:
        start_csv(csvfile, fields)
        
        self.fxt = CsvFixture(  filename=open(csvfile, 'a'), fields=fields, 
                                addfields=True, data=data, clean=True)
        self.fxt.close()
        #print readfile(csvfile, pad=True)
    
        c = csv.DictReader(open(csvfile, 'r'))
        for row in c:
            if row['id'] == 'id':
                assert row['name'] == 'name'
                assert row['hair_color'] == 'hair_color'
            if row['id'] == '1':
                assert row['name'] == self.fxt.blue_hair_color_employee.name
                assert (row['hair_color'] == 
                        self.fxt.blue_hair_color_employee.hair_color)
            elif row['id'] == '2':
                assert row['name'] == self.fxt.boring_employee.name
                assert (row['hair_color'] == 
                        self.fxt.boring_employee.hair_color)
            else:
                raise AssertionError('unexpected row: %s' % row)
    
    def test_affix(self):
        fields = ['id', 'name', 'hair_color']
        csvfile = TMP.mkfile('employee.csv')
        
        self.fxt = CsvFixture(  
                    filename=open(csvfile, 'w'), fields=fields, 
                    addfields=False, data = EmployeeData.meta.data)
        self.fxt.close()
        
        c = csv.DictReader(open(csvfile, 'r'), fieldnames=fields)
        for row in c:
            if row['id'] == '1':
                assert row['name'] == self.fxt.cool_employee.name
                assert row['hair_color'] == self.fxt.cool_employee.hair_color
            elif row['id'] == '2':
                assert row['name'] == self.fxt.nerd_employee.name
                assert row['hair_color'] == self.fxt.nerd_employee.hair_color
            else:
                raise AssertionError('unexpected row: %s' % row)
    
    def test_csv_dict(self):
        fields = ["Foo's field", "BAR BAR"]
        csvfile = TMP.mkfile('csvdict.csv')
        
        self.fxt = Fixture()
        self.fxt = affix(csvdict(filename=csvfile, fields=fields, 
                        addfields=True))
        self.fxt.odd_fields
        
        c = csv.DictReader(open(csvfile, 'r'), fieldnames=fields)
        num = 0
        for row in c:
            if num == 0:
                # assert first line is fields :
                assert row["Foo's field"] == "Foo's field"
                assert row["BAR BAR"] == "BAR BAR"
            elif num == 1:
                self.fxt.assert_has(row)
            num += 1
        self.assertEqual(num, 2)
    
    def test_lazy_fields_from_meta(self):
        class Products(CsvFixture):
            class meta:
                fields = None
                addfields = True
                data = (('truck', dict(color='red', price='10.00')), 
                        ('car', dict(color='blue', price='5.00')),)
                        
        csvfile = TMP.mkfile('csvdict_meta.csv')
        self.fxt = Products(filename = csvfile)
        self.fxt.close()
        
        c = csv.DictReader(open(csvfile, 'r'), 
                        fieldnames=self.fxt.meta.fields)
        num = 0
        for row in c:
            if num == 0:
                [self.assertEqual(row[f], f) for f in self.fxt.meta.fields]
            else:
                self.fxt.assert_has(row)
            num += 1
        self.assertEqual(num, 3)
        
    def test_lazy_fields_from_def(self):
        class Products(CsvFixture):
            class meta:
                fields = None
                addfields = True
            def data(self):
                return (('truck', dict(color='red', price='10.00')), 
                        ('car', dict(color='blue', price='5.00')),)
                        
        csvfile = TMP.mkfile('csvdict_def.csv')
        self.fxt = Products(filename = csvfile)
        self.fxt.close()
        
        c = csv.DictReader(open(csvfile, 'r'), 
                    fieldnames=self.fxt.meta.fields)
        num = 0
        for row in c:
            if num == 0:
                [self.assertEqual(row[f], f) for f in self.fxt.meta.fields]
            else:
                self.fxt.assert_has(row)
            num += 1
        self.assertEqual(num, 3)
    
    def test_encoding(self):
        '''assert fixture vals can be in unicode ...'''
        fields = [ 'Price', 'Page' ]
        data = (('russell_tshirt', { 'Price': u'\xa33.68', 'Page': u'57',}),)
        
        csvfile = TMP.mkfile('test_fxt_unicode.csv')
        self.fxt = CsvFixture(  filename=open(csvfile, 'w'), fields=fields, 
                                addfields=False, encoding='utf-8', data=data)
        self.fxt.close()
        
        #print readfile(csvfile)    
        c = csv.DictReader(open(csvfile, 'r'), fieldnames=fields)
        found = False
        for row in c:
            self.assertEqual(   row['Price'].decode('utf-8'), 
                                self.fxt.russell_tshirt.Price)
            self.assertEqual(   row['Page'].decode('utf-8'), 
                                self.fxt.russell_tshirt.Page)
            found = True
        assert found
    
    def test_on_reload_from_registry(self):
        # when we reload from registry (should registry be allowed here?)
        # we need to recreate the file if it doesn't exist
        class MyUniqueCsv(CsvFixture):
            class meta:
                fields = None
                addfields = True
            def data(self):
                return (('say_hello', dict(hello='bonjour')),)
        
        thefile = open(path.join(TMP.root, 'test_unique.csv'), 'w')
        f = MyUniqueCsv(filename = thefile)
        f.close()
        assert path.exists(thefile.name)
        
        contents = readfile(thefile.name)
        os.unlink(thefile.name)
        
        thefile = open(path.join(TMP.root, 'test_unique.csv'), 'w')
        f = MyUniqueCsv(filename = thefile)
        f.close()
        assert path.exists(thefile.name)
        self.assertEqual(readfile(thefile.name), contents)

    def test_csv_fields_are_in_order(self):
        csvfile = TMP.mkfile('fields_in_order.csv')
        
        class MyCsv(CsvFixture):
            class meta:
                fields = ['first_field', 'second_field', 'third_field']
                addfields = True
            def data(self):
                return (
                ('one', dict(
                    second_field=10, first_field='monday', 
                    third_field='happy')),
                ('two', dict(
                    first_field='tuesday', second_field=8, 
                    third_field='moody'))
                    )
        
        myc = MyCsv(filename = open(csvfile, 'w'))
        myc.close()

        cfile = csv.reader(open(csvfile))
        hdr = cfile.next()

        assert hdr == MyCsv.meta.fields
    
    def test_unknown_meta_keywords(self):
        class BadCsv(CsvFixture):
            class meta:
                fieldnames = ('blah1','blah2')
            def data(self):
                return (('chatty', dict(blah1='umm', blah2='bah')),)
        
        self.assertRaises(TypeError, BadCsv, 
                filename = open(TMP.mkfile('badmeta.csv'), 'w'))
    
    def test_mknewfile(self):
        # mknewfile is a file factory that should 
        # return an open fileobj :
        
        @metamethod
        def mknewfile2(fxt):
            self.assertEqual( fxt.__class__.__name__, 'MyAutoCsv')
            return open(TMP.mkfile('laurel_hardy.csv'), 'w')
        
        # fyi : mknewfile=mknewfile is tripping up nosetests somehow
        class MyAutoCsv(CsvFixture):
            class meta:
                fields = ['laurel','hardy']
                mknewfile = mknewfile2
            def data(self):
                return (('gag', dict(laurel="hammer", hardy="frying pan")),)
        
        fxt = MyAutoCsv()
        fxt.close()
        assert path.exists(path.join(TMP.root, 'laurel_hardy.csv'))
        
        # spot check to make sure referencing doesn't screw anything up
        # when it's added as a keyword :
        @metamethod
        def mknewfile_kw(fxt):
            return open(TMP.mkfile('laurel_hardy2.csv'), 'w')
            
        fxt = CsvFixture(
                mknewfile = mknewfile_kw,
                data = (('noop', dict(ping='pong')),))
        fxt.close()
        assert path.exists(path.join(TMP.root, 'laurel_hardy2.csv'))

class TestSOFixture(unittest.TestCase):
    def setUp(self):
        model.Employee.clearTable()
        LoadMap.clear()
        reset_debug()
        
    def test_load(self):
        fxt = SOFixture(data=(('bob', {'name':'bob','hair_color':'black'}),), 
                        so_class = model.Employee)
        
        rs = model.Employee.select()
        assert rs.count()
        for emp in rs:
            fxt.assert_has(emp, transkey={'hair_color':'hairColor'})
    
    def test_data_as_meta(self):
        class Employee(SOFixture):
            class meta:
                env = model
                data = (('jane', {'name':'jane','hair_color':'brown'}),)
        fxt = Employee()
        
        rs = model.Employee.select()
        assert rs.count()
        for emp in rs:
            fxt.assert_has(emp, transkey={'hair_color':'hairColor'})
    
    def test_data_and_so_class_as_meta(self):
        class EmployeeNoMagic(SOFixture):
            class meta:
                so_class = model.Employee
                data = (('jane', {'name':'jane','hair_color':'brown'}),)
        fxt = EmployeeNoMagic()
        
        rs = model.Employee.select()
        assert rs.count()
        for emp in rs:
            fxt.assert_has(emp, transkey={'hair_color':'hairColor'})
    
    def test_data_callable(self):
        class Employee(SOFixture):
            class meta:
                env = model
            def data(self):
                return (('jane', {'name':'jane','hair_color':'brown'}),)
                
        fxt = Employee()
        
        rs = model.Employee.select()
        assert rs.count()
        for emp in rs:
            fxt.assert_has(emp, transkey={'hair_color':'hairColor'})
    
    def test_data_callable_nomagic(self):
        class EmployeeNoMagic(SOFixture):
            class meta:
                so_class = model.Employee
            def data(self):
                return (('jane_doe', {'name':'jane','hair_color':'brown'}),)
                
        fxt = EmployeeNoMagic()
        fxt.jane_doe
        
        rs = model.Employee.select()
        assert rs.count()
        for emp in rs:
            fxt.assert_has(emp, transkey={'hair_color':'hairColor'})
    
    def test_data_wrong_ref(self):
        
        # was getting the wrong ref here (bob was carrying 
        # through to subclasses)
        fxt = SOFixture(data=(('bob', {'name':'bob','hair_color':'black'}),), 
                        so_class=model.Employee)
        fxt.bob
        class JaneEmployee(SOFixture):
            class meta:
                so_class = model.Employee
            def data(self):
                return (('jane', {'name':'jane','hair_color':'brown'}),)
        fxt = JaneEmployee()
        fxt.jane

    def test_dependencies(self):
    
        old = set_debug_stream(sys.stdout)
    
        unregister_fixtures()
        model.Employee.clearTable()
        model.Department.clearTable()
    
        fxt = DepartmentData()
    
        assert model.Department.select().count() > 0
        # problem is that DepartmentData is somehow getting its 
        # meta.data attribute set
        assert model.Employee.select().count() > 0

        fxt.clean()

        assert model.Department.select().count() == 0
        assert model.Employee.select().count() == 0
        set_debug_stream(old)
    
    def test_find_so_class(self):
            
        model.Department.clearTable()
        model.Shop.clearTable()
        model.Employee.clearTable()
        
        fxt = EmployeeData()
        self.assertEqual( fxt.meta.so_class, model.Employee)
        fxt = DepartmentData()
        self.assertEqual( fxt.meta.so_class, model.Department)
        fxt = ShopData()
        self.assertEqual( fxt.meta.so_class, model.Shop)
        
    def test_lazy_load(self):
        from sqlobject import connectionForURI, SQLObject, ForeignKey, StringCol
        class Base:
            _connection = connectionForURI('sqlite:/:memory:')
        
        class Clients(Base, SQLObject):
            name = StringCol()
        class ClientData(SOFixture):
            class meta:
                so_class = Clients
            def data(self):
                return (('acme', dict(id=9, name='acme')),)
            
        class Products(Base, SQLObject):
            client = ForeignKey('Clients')
            name = StringCol()
        class ProductData(SOFixture):
            class meta:
                so_class = Products
            def data(self):
                return (('anvil', dict(id=8, client_id=ClientData().acme.id, 
                                        name='anvil')),)
            
        class Offers(Base, SQLObject):
            product = ForeignKey('Products')
            name = StringCol()
        class OfferData(SOFixture):
            class meta:
                so_class = Offers
            def data(self):
                return (('anvil_sale', dict(id=7, name='half price anvils', 
                                        product_id=ProductData().anvil.id)),)
        Clients.createTable()
        Products.createTable()
        Offers.createTable()
    
        # sqlite doesn't support foreign key constraints, 
        # so using assertions for that ...
    
        offers = OfferData()
        rs = offers.meta.so_class.select()
        assert rs.count()==1
    
        # make sure other tables lazy loaded :
        self.assertEqual( Products.select().count(), 1)
        self.assertEqual( Clients.select().count(), 1)
    
        products = ProductData()
        clients = ClientData()
    
        row = rs[0]
        assert row.name == offers.anvil_sale.name
        assert row.productID == products.anvil.id
    
        rs = products.meta.so_class.select()
        assert rs.count()==1
        row = rs[0]
        assert row.name == products.anvil.name
        assert row.clientID == clients.acme.id
    
        rs = clients.meta.so_class.select()
        assert rs.count()==1
        assert rs[0].name == clients.acme.name
    
        # make sure nothing double loaded:
        assert Offers.select().count()==1
        assert Products.select().count()==1
        assert Clients.select().count()==1
    
    def test_transactions(self):
        # we actually can't run this test with sqlite.
        # that is really annoying
        pass
        # trans = model.Employee._connection.transaction()
        # 
        # class EmployeeData(SOFixture):
        #     class meta:
        #         clean = True
        #         so_class = model.Employee
        #     def data(self):
        #         return (('jane_doe', {'name':'jane','hair_color':'brown'}),)
        #         
        # fxt = EmployeeData()
        # self.assertEqual(model.Employee.select(connection = trans).count(), 0)
        # fxt.close()
        # self.assertEqual(model.Employee.select(connection = trans).count(), 1)
    
    def test_transactions_autocommit(self):
        # can use autocommit keyword to workaround postgresql deadlocks.
        # arrrrggh.  Note that this can never actually fail when using sqlite.
        
        class EmployeeData(SOFixture):
            class meta:
                clean = True
                autocommit = True
                so_class = model.Employee
            def data(self):
                return (('jane_doe', {'name':'jane','hair_color':'brown'}),)
                
        self.assertEqual(model.Employee.select().count(), 0)
        fxt = EmployeeData()
        # it must exist before close() :
        self.assertEqual(model.Employee.select().count(), 1)
        fxt.close()
    
    def test_transactions_sqlite_bug(self):
        # test that saves are performed within a transaction
        # since sqlite is not threaded and *could* already 
        # be in a transaction.
        
        # start any kind of transaction
        trans = model.Employee._connection.transaction()
        empty_rs = model.Employee.selectBy(name='bob', hairColor='blk', 
                                            connection=trans)
        trans.commit()
        
        class EmployeeData(SOFixture):
            class meta:
                clean = True
                so_class = model.Employee
            def data(self):
                return (('jane_doe', {'name':'jane','hair_color':'brown'}),)
        
        # bug in sqlobject makes it so that if a transaction
        # was used for *any* other part of the program
        # then sqlobject will never go back to autocommit mode
        fxt = EmployeeData()
        self.assertEqual(model.Employee.select().count(), 1)
        fxt.close()
        
        # do another transaction to make sure the previous one 
        # didn't *bleed* into this one (sqlite bug/ quirk)
        trans = model.Employee._connection.transaction()
        bob = model.Employee(name='bob', hairColor='blk', connection=trans)
        trans.rollback()
        self.assertEqual(model.Employee.select().count(), 1)

class TestSOFinders:
    def setUp(self):
        class meta:
            pass
        self.meta = meta

    def test_find_so_from_data(self):
        assert find_so_from_data(self.meta(), 'ProductsData') == "Products"
        
    def test_find_so_from_fixture(self):
        assert find_so_from_fixture(self.meta(), 'Products') == "Products"
        
    def test_find_so_from_table(self):
        assert find_so_from_table(self.meta(), 'products') == "Products"
        assert (find_so_from_table(self.meta(), 'products_to_categories') == 
                                                    "ProductsToCategories")
        assert find_so_from_table(self.meta(), 'products_id') == "ProductsID"
